# CompuCanvas
# Digital Simulation
# Written by Linda and Scott Wills
# (c) 2009-2011 Scott & Linda Wills

# This code performs digital simulation of connected objects.
# Objects are positioned on a regular gridded contact board.
# Objects contain terminals that are connected by wire and bars,
# labels (including vdd and gnd) and contact colocation.
# Multiple terminals connect to form nodes. Because switches
# are also # simulated, terminals are not always designated as
# input or output. Rather object simulation can return
# temporarily connected nodes (cnode). Labeled contacts
# connected to node can also permanently join nodes
#
# TODO: to improve implementation, replace lists with sets

class Node :
    def __init__(Self, Parent=None, Terminals=[]) :
        Self.Parent = Parent           # link to simulation
        Self.Terminals = Terminals     # terminal list
        Self.State = 5                 # initial state

class Simulation :
    def __init__(Self, Parent=None) :
        Self.Parent = Parent           # link to simulation
        Self.Nodes = []                # list of nodes
        Self.Cnodes = []               # list of switch connected nodes
        Self.Contacts = []             # contact point positions
        Self.Labels = []               # labeled contacts

    def Derive_Nodes(Self, Objects) :
        """ This routine derives and returns the nodes using the the
        object list. It begins by creating a contact dictionary, label
        dictionary, and link list. The contact dictionary contains
        list of colocated terminals index by the terminal position.
        The label dictionary contains a list of connected terminals
        based on label name (including VDD and GND). The Links list
        contains lists of terminals connected by wires and bars. All
        of these collections are permanently connected terminals. Then
        these sets are merged into a collection of nodes containing
        terminal lists. """
        Sets = []
        Contacts = {}
        Labels = {}
        Links = []
        for Object in Objects :
            for Terminal in Object.Terminals :
                if Terminal.Active :
                    C = tuple(map(sum, zip(Terminal.Contact, Object.Base)))
                    if C in Contacts :
                        Contacts[C].append(Terminal)
                    else :
                        Contacts[C] = [Terminal]
            if Object.Type in ('LABEL', 'VDD', 'GND') :
                if Object.Text in Labels :
                    Labels[Object.Text].append(Object.Terminals[0])
                else :
                    Labels[Object.Text] = [Object.Terminals[0]]
            if Object.Type in ('WIRE', 'BAR') :
                Links.append(Object.Terminals)

        # All permanent connection collections are now created

        Sets = Contacts.values()              # node sets begin with colocated contact lists
        for Terminals in Labels.values() :
            if len(Terminals) > 1 :           # if terminals connected via symbolic labels
                Sets = Self.Merge_Sets(Terminals, Sets) # merge sets connected by list members
        for Terminals in Links :              # if terminals connected via wires and bars
            Sets = Self.Merge_Sets(Terminals, Sets)  # merge sets connected by list members

        # Sets now contains all permanently connected terminals

        Self.Nodes = []                       # create new node list
        for Set in Sets :                     # create a new node for each set of connected terminals
            NewNode = Node(Self, Set)
            Self.Nodes.append(NewNode)
            for Terminal in Set :             # cross link terminals and nodes
                Terminal.Node = NewNode
        Self.Labels = Labels.keys()           # create labels list
        Self.Contacts = Contacts.keys()       # create contacts list

    def Merge_Sets(Self, List, Sets) :
        """ This routine processes a list of connected terminals to
        compute the transitive closure affect of already connected
        terminal lists in Sets. Sets are non-overlapping and a terminal
        appears only once in sets. The sets is extracted and merged,
        then reinserted into Sets as a combined grouping."""
        NewSet = []
        for Element in List :               # for each newly connected terminal
            if Element not in NewSet :      # if not already in aggregrate collection
                Set = [S for S in Sets if Element in S][0] # find a set that contains the terminal
                Sets.remove(Set)            # remove for sets of connected terminals
                NewSet += Set               # add matched set to new aggregrate collection
        Sets.append(NewSet)                 # when all matching sets as aggretated, add new set
        return Sets                         # return updated sets

    def Simulate_Objects(Self, Objects) :
        """ This routine simulates each object based on the current node states.
        Terminal sources are updated. and a new Cnode list is created. """
        Self.Parent.Menu.Clear_Indicators()
        Self.Cnodes = []
        for Node in Self.Nodes :
            Self.Cnodes.append([Node])
        for Object in Objects :
            Cnode = Object.Simulate()
            if Cnode :
                Self.Cnodes = Self.Merge_Sets(Cnode, Self.Cnodes)

    def Find_Node_State(Self) :
        """ This routine determines the state of each node. It processes each cnode and,
        based on connected sources, determines the new state of each node in the collection. """
        if not Self.Cnodes :
            Self.Cnodes = map(list, zip(Self.Nodes))
        for Cnode in Self.Cnodes :
            Sources = []
            for Node in Cnode :
                for Terminal in Node.Terminals :
                    if Terminal.Source not in Sources :
                        Sources.append(Terminal.Source)
            if '0' in Sources and '1' in Sources :
                State = 4      # short
                Self.Parent.Menu.Indicators["Short"].Update(True)
            elif '1' in Sources :
                State = 1      # vdd
            elif '0' in Sources :
                State = 0      # gnd
            elif 'X' in Sources :
                State = 2      # unknown
            else :
                State = 3      # float
            for Node in Cnode :
                Node.State = State
