# CompuCanvas
# Objects and Terminals
# Written by Linda and Scott Wills
# (c) 2009-2011 Scott & Linda Wills

from copy           import copy
from Tkinter        import ARC, TOP, X, LEFT, RIGHT, YES
from Tkinter        import Toplevel, IntVar, StringVar, Frame, Label, Radiobutton, Button, Entry

############################### Terminal ##############################

class Terminal () :
    """ This is the primary terminal class. A terminal include a connection
    point offset, a stub point offset, and a parent object. When bubbled, a
    correction B[] is computed to shorten the stub. """
    def __init__(Self, Object, Contact, Stub, Active=True, Bubble=False, Source="F") :
        Self.Object = Object
        Self.Contact = Contact
        Self.Stub = Stub
        Self.Active = Active
        Self.Bubble = Bubble
        Self.Source = Source
        Self.Node = None

    def Draw(Self) :
        """ This routine draws a terminal stub. """
        if Self.Active :
            C = copy(Self.Contact)
            S = copy(Self.Stub)
            for I in (0, 1) :
                C[I] = Self.Contact[I] + Self.Object.Base[I]
                if Self.Bubble :
                    if Self.Contact[I] > Self.Stub[I] :
                        BubbleCorrection = 10
                    elif Self.Contact[I] < Self.Stub[I] :
                        BubbleCorrection = -10
                    else :
                        BubbleCorrection = 0
                    S[I] = Self.Stub[I] + BubbleCorrection + Self.Object.Base[I]
                else :
                    S[I] = Self.Stub[I] + Self.Object.Base[I]
            if Self.Node :
                Color = Self.Object.Canvas.Colors[Self.Node.State]
            else :
                Color = 'black'
            StubID = Self.Object.Canvas.Create_Scaled_Line(C[0], C[1], S[0], S[1], fill=Color, tag=Self.Object.ID)
            Self.Object.Canvas.tag_lower(StubID, Self.Object.ID)

######################### CompuCanvas Objects #########################

IDmax = 0

def NextID() :
    """ This routine returns the next ID. """
    global IDmax
    IDmax += 1
    return IDmax

class Object() :
    """ This is a base compucanvas object. """
    def __init__(Self, Canvas, Base=[0,0]) :
        Self.Canvas = Canvas
        Self.Base = Base
        Self.HighlightID = None

    def Move(Self, Delta) :
        """ This routine moves the object by the specified grid delta. """
        Self.Base[0] += 10*Delta[0]
        Self.Base[1] += 10*Delta[1]
        Self.Canvas.Scaled_Move(Self.ID, 10 * Delta[0], 10 * Delta[1])

    def Highlight(Self, On) :
        """ This routine highlights or unhighlights an object. """
        if On :
            X1, Y1 = Self.Base
            X2 = X1 + Self.Size[0]
            Y2 = Y1 + Self.Size[1]
            ID = Self.ID
            Self.HighlightID = Self.Canvas.Create_Scaled_Rectangle(X1, Y1, X2, Y2, width=0, fill="lightgray", tag=Self.ID)
            Self.Canvas.tag_lower(Self.HighlightID, ID)
        else :
            Self.Canvas.delete(Self.HighlightID)
            Self.HighlightID = None

    def Erase(Self) :
        """ This routine erases the object from the canvas. """
        Self.Canvas.delete(Self.ID)

    def Properties(Self, ObjList) :
        """ This routine provides the default properties behavior for objects
        without modifiable properties. It skips menu building and selection
        and continues the object list call chain. """
        Self.Update_Properties(None, None, ObjList)

    def Update_Properties(Self, Vars, Win, ObjList) :
        """ This routine implements a call chain for properties. The updating
        of properties (if necessary) is performed elsewhere. If Vars are provided
        and the object list is not empty, the next object is called. Otherwise
        the call chain terminates by calling Properties_Complete. """
        Self.Highlight(False)
        if Win :
            Win.destroy()
        if (Vars <> []) and ObjList :
            ObjList[0].Properties(ObjList[1:])
        else :
            Self.Canvas.master.Properties_Complete()

    def Within(Self, Point) :
        """ This routine determines whether a point is within the object's
        bounding box. """
        Xp, Yp = Point
        Xmin, Ymin = Self.Base
        Xmax, Ymax = Xmin + Self.Size[0], Ymin + Self.Size[1]
        return Xmin <= Xp <= Xmax and Ymin <= Yp <= Ymax

    def Simulate(Self) :
        """ This routine simulates a non-behavioral object (i.e., it has no
        explicit behavior, rather it is a connector or source. """
        return []

    def Write(Self, File) :
        """ This routine writes an object to the output file. """
        File.write("%s(Base=%s)\n" % (Self.Type, Self.Base))

    def Clone(Self) :
        """ This routine returns a clone string. """
        return "%s(Self.Canvas, Base=%s)" % (Self.Type, Self.Base)

class PASS(Object) :
    def __init__(Self, Canvas, Base=[0,0]) :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'PASS'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (60, 60)
        Self.Switch = 0
        Self.Terminals = (Terminal(Self, [30,0], [30,10]),
                          Terminal(Self, [30,60], [30,40]),
                          Terminal(Self, [0,30], [10,30]),
                          Terminal(Self, [60,30], [50,30]))

    def Simulate(Self) :
        """ This routine simulates a pass gate. If the control inputs
        open both switches, the switch terminals are disconnected. If
        either switch is closed, the switched terminals are connected.
        However bad inputs and unmatched switch controls will activate
        the canvas warning. A Cnode list is returned if the switch is
        closed. """
        Self.Switch = 0
        BadIn = False
        for Terminal in Self.Terminals[:2] :
            if Terminal.Node.State == Self.Terminals.index(Terminal) :
                Self.Switch += 1
            elif Terminal.Node.State > 1 :
                BadIn = True
        if BadIn or Self.Switch == 1 :
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
        if Self.Switch > 0 :
            return [Self.Terminals[2].Node, Self.Terminals[3].Node]
        else :
            return []

    def Draw(Self) :
        """ This routine draws a PASS gate. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Line(X+10, Y+10, X+10, Y+50, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+50, Y+10, X+50, Y+50, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+10, X+50, Y+30, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+30, X+50, Y+10, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+30, X+50, Y+50, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+50, X+50, Y+30, tag=ID)
        Self.Canvas.Create_Scaled_Oval(X+25, Y+10, X+35, Y+20, tag=ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()
        if Self.Canvas.master.SwitchMarker :
            if Self.Switch == 1 :
                Self.Canvas.Create_Scaled_Line(X+15, Y+30, X+45, Y+30, fill="cyan", dash = (2), tag=ID)
            elif Self.Switch == 2 :
                Self.Canvas.Create_Scaled_Line(X+15, Y+30, X+45, Y+30, fill="cyan", tag=ID)

class DOT(Object) :
    def __init__(Self, Canvas, Base=[0,0]) :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'DOT'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (0,0)
        Self.Terminals = (Terminal(Self, [0,0], [0,0]),)

    def Draw(Self) :
        """ This routine draws a dot object. """
        X, Y = Self.Base
        if Self.Terminals[0].Node :
            Color = Self.Canvas.Colors[Self.Terminals[0].Node.State]
        else :
            Color = 'black'
        Self.Canvas.Create_Scaled_Oval(X-2, Y-2, X+2, Y+2, outline=Color, fill=Color, tag=Self.ID)

    def Highlight(Self, On) :
        """ This routine highlights or unhighlights a dot object. """
        if On :
            X, Y = Self.Base
            ID = Self.ID
            Self.HighlightID = Self.Canvas.Create_Scaled_Oval(X-4, Y-4, X+4, Y+4, width=0, fill="lightgray", tag=ID)
            Self.Canvas.tag_lower(Self.HighlightID, ID)
        else :
            Self.Canvas.delete(Self.HighlightID)
            Self.HighlightID = None

class VDD(Object) :
    def __init__(Self, Canvas, Base=[0,0]) :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'VDD'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (40, 10)
        Self.Terminals = (Terminal(Self, [20,10], [20,0], Source='1'),)
        Self.Text = 'VDD'

    def Draw(Self) :
        """ This routine draws a VDD supply. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Line(X, Y, X+40, Y, tag=ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()

class GND(Object) :
    def __init__(Self, Canvas, Base=[0,0]) :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'GND'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (40, 30)
        Self.Terminals = (Terminal(Self, [20,0], [20,10], Source='0'),)
        Self.Text = 'GND'

    def Draw(Self) :
        """ This routine draws a GND supply. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Line(X, Y+10, X+40, Y+10, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+20, X+30, Y+20, tag=ID)
        Self.Canvas.Create_Scaled_Oval(X+20, Y+30, X+20, Y+30, tag=ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()

class LABEL(Object) :
    def __init__(Self, Canvas, Base=[0,0], Text=None, Anchor='W', Source='F') :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'LABEL'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (0,0)
        Self.Terminals = (Terminal(Self, [0,0], [0,0], Source=Source),)
        if Text :
            Self.Text = Text
        else :
            Self.Text = Self.Canvas.master.Next_Label()
        Blocks = ""
        for Obj in Self.Canvas.master.Objects :   # heuristic anchor position
            Xmin, Ymin = Obj.Base
            Xmax, Ymax = map (lambda X, Y : X + Y, Obj.Base, Obj.Size)
            if Xmin <= Base[0] <= Xmax and Ymin <= Base[1] <= Ymax :
                if Base[0] == Xmin and 'E' not in Blocks :
                    Blocks += "E"
                elif Base[0] == Xmax and 'W' not in Blocks :
                    Blocks += "W"
                elif Base[1] == Ymin and 'S' not in Blocks :
                    Blocks += "S"
                elif Base[1] == Ymax and 'N' not in Blocks :
                    Blocks += "N"
        if len(Blocks) == 0 :
            Self.Anchor = Anchor
        elif len(Blocks) == 1 or len(Blocks) == 4 :
            Self.Anchor = "WESN"["EWNS".find(Blocks[0])]
        else :
            for Anchor in "SNEW" :
                if Anchor not in Blocks :
                    Self.Anchor = Anchor
        Self.Offsets = {'E': ('w', (2,0)), 'W': ('e', (-2,0)), 'N': ('s', (0,-2)), 'S': ('n', (0,2))}

    def Properties(Self, ObjList) :
        """ This routine builds a label property menu that includes
        the label string, anchor, and font size. """
        Self.Highlight(True)
        PropertiesForm = Toplevel()
        PropertiesForm.title('Properties')

        Text = StringVar()
        Text.set(Self.Text)
        TextFrame = Frame(PropertiesForm)
        TextFrame.pack(side=TOP, fill=X)
        TextLabel = Label(TextFrame, text='label', width=15)
        TextLabel.pack(side=LEFT)
        TextEntry = Entry(TextFrame)
        TextEntry.config(textvariable=Text)
        TextEntry.pack(side=RIGHT, expand=YES, fill=X)

        Anchor = StringVar()
        Anchor.set(Self.Anchor)
        AnchorFrame = Frame(PropertiesForm)
        AnchorFrame.pack(side=TOP, fill=X)
        AnchorLabel = Label(AnchorFrame, text='position', width=15)
        AnchorLabel.pack(side=LEFT)
        for Pos in ('E', 'N', 'W', 'S') :
            AnchorEntry = Radiobutton(AnchorFrame, text=Pos, variable=Anchor, value=Pos)
            AnchorEntry.pack(side=LEFT)

        Source = StringVar()
        Source.set(Self.Terminals[0].Source)
        SourceFrame = Frame(PropertiesForm)
        SourceFrame.pack(side=TOP, fill=X)
        SourceLabel = Label(SourceFrame, text='source', width=15)
        SourceLabel.pack(side=LEFT)
        for Nam, Src in (('1', '1'), ('0', '0'), ('let go', 'F')) :
            SourceEntry = Radiobutton(SourceFrame, text=Nam, variable=Source, value=Src)
            SourceEntry.pack(side=LEFT)

        Parameters = (Text, Anchor, Source)
        ButtonFrame = Frame(PropertiesForm)
        ButtonFrame.pack(side=TOP, fill=X)
        OKButton = Button(ButtonFrame, text='OK', command=(lambda Vars=Parameters,
                                                           Win=PropertiesForm,
                                                           List=ObjList:
                                                           Self.Update_Properties(Vars, Win, List)))
        OKButton.pack(side=LEFT, expand=YES, fill=X)
        CancelButton = Button(ButtonFrame, text='Cancel',  command=(lambda Vars=[],
                                                                    Win=PropertiesForm,
                                                                    List=ObjList:
                                                                    Self.Update_Properties(Vars, Win, ObjList)))
        CancelButton.pack(side=RIGHT, expand=YES, fill=X)

    def Update_Properties(Self, Vars, Win, ObjList) :
        """ This routine updates label properties if a list is provided.
        After new parameters are processed (or if they are not provided),
        the object update_properties routine is called to implement the
        object list call chain. """
        if Vars :
            if Vars[0].get() <> Self.Text :
                Self.Text = Vars[0].get()
                Source = Self.Get_Source()
                if Source and Vars[2].get() == Self.Terminals[0].Source :
                    Vars[2].set(Source)
            if Vars[1].get() <> Self.Anchor :
                Self.Anchor = Vars[1].get()
            if Vars[2].get() <> Self.Terminals[0].Source :
                Self.Set_Source(Vars[2].get())
                Self.Canvas.master.Update_State()
        Object.Update_Properties(Self, Vars, Win, ObjList)

    def Draw(Self, Bold=False) :
        """ This routine draws a label. By keyword, the face can be bold.
        The text ID is returned. """
        Anchor, Offset = Self.Offsets[Self.Anchor]
        X = Self.Base[0] + Offset[0]
        Y = Self.Base[1] + Offset[1]
        if Self.Text[0] == '-' :
            Text = Self.Text[1:]
            if Bold :
                Font = Self.Canvas.Bold
            else :
                Font = Self.Canvas.Font
            W = Font.measure(Text) / Self.Canvas.master.Scale
            H = 4 + Font.metrics("ascent") / Self.Canvas.master.Scale
            A = 4 / Self.Canvas.master.Scale
            if Anchor == 'e' :
                Self.Canvas.Create_Scaled_Line(X-W, Y-H/2, X, Y-H/2, tag=Self.ID)
            elif Anchor == 'w' :
                Self.Canvas.Create_Scaled_Line(X, Y-H/2, X+W, Y-H/2, tag=Self.ID)
            elif Anchor == 's' :
                Self.Canvas.Create_Scaled_Line(X-W/2, Y-H-A, X+W/2, Y-H-A, tag=Self.ID)
            elif Anchor == 'n' :
                Self.Canvas.Create_Scaled_Line(X-W/2, Y, X+W/2, Y, tag=Self.ID)
        else :
            Text = Self.Text
        if Bold :
            return Self.Canvas.Create_Scaled_Text(X, Y, text=Text, anchor=Anchor, font=Self.Canvas.Bold, tag=Self.ID)
        else :
            return Self.Canvas.Create_Scaled_Text(X, Y, text=Text, anchor=Anchor, font=Self.Canvas.Font, tag=Self.ID)

    def Highlight(Self, On) :
        """ This routine highlights or unhighlights a label object. """
        if On :
            Self.Canvas.delete(Self.ID)
            Self.HighlightID = Self.Draw(Bold=True)
        else :
            Self.Canvas.delete(Self.ID)
            Self.HighlightID = None
            Self.Draw()

    def Toggle_Source(Self) :
        """ This routine toggle the connected source. Sources '0' and
        '1' toggle. If the source is 'F', it toggles to zero. """
        if Self.Terminals[0].Source == '0' :
            Self.Set_Source('1')
        elif Self.Terminals[0].Source in ['1', 'F'] :
            Self.Set_Source('0')
        Self.Canvas.master.Update_State()

    def Get_Source(Self) :
        """ This routine gets a source from a matching label. If no labels
        match, the returned value is None."""
        for Object in Self.Canvas.master.Objects :
            if Object.Type == 'LABEL' and Object.Text == Self.Text :
                return (Object.Terminals[0].Source)
        return None

    def Set_Source(Self, Source) :
        """ This routine sets a label source to a given value. Because many
        label objects may share the same name, all label objects are checked
        and updated if appropriate. """
        for Object in Self.Canvas.master.Objects :
            if Object.Type == 'LABEL' and Object.Text == Self.Text :
                Object.Terminals[0].Source = Source

    def Write(Self, File) :
        """ This routine writes a label to the output file. """
        File.write("%s(Base=%s, Text= '%s', Anchor= '%s', Source= '%s')\n" % \
                   (Self.Type, Self.Base, Self.Text, Self.Anchor, Self.Terminals[0].Source))

    def Clone(Self) :
        """ This routine returns a clone string. """
        return "%s(Self.Canvas, Base=%s, Text= '%s', Anchor= '%s', Source= '%s')" % \
               (Self.Type, Self.Base, Self.Text, Self.Anchor, Self.Terminals[0].Source)

class CLOCK(Object) :
    def __init__(Self, Canvas, Base=[0,0], Phase=1, Depth=5, Delay=1) :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'CLOCK'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (40,80)
        Self.Terminals = (Terminal(Self, [20,0], [20,10]),)
        Self.Phase = Phase
        Self.Depth = Depth
        Self.Delay = Delay

    def Properties(Self, ObjList) :
        """ This routine builds a label property menu that includes
        the phase, phase depth, and non-overlap delay. """
        PropertiesForm = Toplevel()
        PropertiesForm.title('Properties')

        Phase = IntVar()
        Phase.set(Self.Phase)
        PhaseFrame = Frame(PropertiesForm)
        PhaseFrame.pack(side=TOP, fill=X)
        PhaseLabel = Label(PhaseFrame, text='phase', width=15)
        PhaseLabel.pack(side=LEFT)
        for Size in xrange(1, 3) :
            PhaseEntry = Radiobutton(PhaseFrame, text=str(Size), variable=Phase, value=Size)
            PhaseEntry.pack(side=LEFT)

        Depth = IntVar()
        Depth.set(Self.Depth)
        DepthFrame = Frame(PropertiesForm)
        DepthFrame.pack(side=TOP, fill=X)
        DepthLabel = Label(DepthFrame, text='phase depth', width=15)
        DepthLabel.pack(side=LEFT)
        for Size in xrange(1, 9) :
            DepthEntry = Radiobutton(DepthFrame, text=str(Size), variable=Depth, value=Size)
            DepthEntry.pack(side=LEFT)

        Delay = IntVar()
        Delay.set(Self.Delay)
        DelayFrame = Frame(PropertiesForm)
        DelayFrame.pack(side=TOP, fill=X)
        DelayLabel = Label(DelayFrame, text='non-overlap delay', width=15)
        DelayLabel.pack(side=LEFT)
        for Size in xrange(1, 4) :
            DelayEntry = Radiobutton(DelayFrame, text=str(Size), variable=Delay, value=Size)
            DelayEntry.pack(side=LEFT)

        Parameters = (Phase, Depth, Delay)
        ButtonFrame = Frame(PropertiesForm)
        ButtonFrame.pack(side=TOP, fill=X)
        OKButton = Button(ButtonFrame, text='OK', command=(lambda Vars=Parameters,
                                                           Win=PropertiesForm,
                                                           List=ObjList:
                                                           Self.Update_Properties(Vars, Win, List)))
        OKButton.pack(side=LEFT, expand=YES, fill=X)
        CancelButton = Button(ButtonFrame, text='Cancel',  command=(lambda Vars=[],
                                                                    Win=PropertiesForm,
                                                                    List=ObjList:
                                                                    Self.Update_Properties(Vars, Win, ObjList)))
        CancelButton.pack(side=RIGHT, expand=YES, fill=X)

    def Update_Properties(Self, Vars, Win, ObjList) :
        """ This routine updates label properties if a list is provided.
        After new parameters are processed (or if they are not provided),
        the object update_properties routine is called to implement the
        object list call chain. """
        if Vars :
            if Vars[0].get() <> Self.Phase :
                Self.Phase = Vars[0].get()
            if Vars[1].get() <> Self.Depth :
                Self.Depth = Vars[1].get()
            if Vars[2].get() <> Self.Delay :
                Self.Delay = Vars[2].get()
        Object.Update_Properties(Self, Vars, Win, ObjList)

    def Draw(Self, Bold=False) :
        """ This routine draws a clock. """
        X, Y = Self.Base
        Self.Canvas.Create_Scaled_Oval(X, Y+10, X+40, Y+50, tag=Self.ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+20, X+10, Y+40, tag=Self.ID)
        Self.Canvas.Create_Scaled_Line(X+20, Y+20, X+20, Y+40, tag=Self.ID)
        Self.Canvas.Create_Scaled_Line(X+30, Y+20, X+30, Y+40, tag=Self.ID)
        Self.Canvas.Create_Scaled_Line(X+20, Y+50, X+20, Y+60, tag=Self.ID)
        if Self.Phase == 1 :
            Self.Canvas.Create_Scaled_Line(X+10, Y+20, X+20, Y+20, tag=Self.ID)
            Self.Canvas.Create_Scaled_Line(X+20, Y+40, X+30, Y+40, tag=Self.ID)
        elif Self.Phase == 2 :
            Self.Canvas.Create_Scaled_Line(X+10, Y+40, X+20, Y+40, tag=Self.ID)
            Self.Canvas.Create_Scaled_Line(X+20, Y+20, X+30, Y+20, tag=Self.ID)     
        Self.Canvas.Create_Scaled_Line(X, Y+60, X+40, Y+60, tag=Self.ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+70, X+30, Y+70, tag=Self.ID)
        Self.Canvas.Create_Scaled_Oval(X+20, Y+80, X+20, Y+80, tag=Self.ID)
        # Self.Canvas.Create_Scaled_Text(X+20, Y+10, text=str(Self.Phase), anchor='n', font=Self.Canvas.Font, tag=Self.ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()

    def Set_Source(Self, Cycle, Depth, Delay) :
        """ This routine computes the current clock state based on the
        simulated cycle number plus phase, depth, and delay. """
        Cycle %= 2 * (Depth + Delay)
        if (Cycle < Depth and Self.Phase == 1) or (Depth + Delay <= Cycle < 2 * Depth + Delay and Self.Phase == 2):
            Self.Terminals[0].Source = '1'
        else :
            Self.Terminals[0].Source = '0'

    def Write(Self, File) :
        """ This routine writes a clock object to the output file. """
        File.write("%s(Base=%s, Phase=%s, Depth=%s, Delay=%s)\n" % (Self.Type, Self.Base, Self.Phase, Self.Depth, Self.Delay))

    def Clone(Self) :
        """ This routine returns a clone string. """
        return "%s(Self.Canvas, Base=%s, Phase=%s, Depth=%s, Delay=%s)\n" % (Self.Type, Self.Base, Self.Phase, Self.Depth, Self.Delay)

class WIRE(Object) :
    def __init__(Self, Canvas, Base=[0,0], Points=[]) :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'WIRE'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (0,0)
        Self.Points = Points
        if Points <> [] :
            Self.Finish()
        else :
            Self.Segments = []          # start of a user built wire

    def Add_Point(Self, Point) :
        """ This routine adds a new point and segment to the wire
        object. """
        Last = Self.Points[-1]
        Segment = Self.Canvas.Create_Scaled_Line(Last[0], Last[1], Point[0], Point[1], tag=Self.ID)
        Self.Segments.append(Segment)
        Self.Points.append(Point)

    def Remove_Point(Self) :
        """ This routine removes the last point and segment from the
        current wire object. It cannot remove the first point. """
        if len(Self.Segments) > 0 :
            Self.Points = Self.Points[:-1]
            LastSegment = Self.Segments[-1]
            Self.Canvas.delete(LastSegment)
            Self.Segments = Self.Segments[:-1]

    def Finish(Self, Point=None) :
        """ This routine finishes a wire by computing the wire's base and size.
        Depending on whether the wire was built manually, a split of another
        wire, or a wire loaded from a file, the points may need to be shifted
        and the existing base adjusted. The first and last points are then
        removed from the points list and turned into stubless terminals. """
        if Point :
            Self.Points.append(Point)
        Delta, Self.Size = Self.Canvas.Find_Base_Size(Self.Points, Shift=True)
        Self.Base = [Self.Base[0] + Delta[0], Self.Base[1] + Delta[1]]
        Self.Terminals = [Terminal(Self, Self.Points[0], Self.Points[0]),
                          Terminal(Self, Self.Points[-1], Self.Points[-1])]
        Self.Points = Self.Points[1:-1]
        Self.Segments = None

    def Draw(Self) :
        """ This routine draws a wire. """
        X, Y = Self.Base
        ID = Self.ID
        Last = Self.Terminals[0].Contact
        if Self.Terminals[0].Node :
            Color = Self.Canvas.Colors[Self.Terminals[0].Node.State]
        else :
            Color = 'black'
        for Point in Self.Points + [Self.Terminals[1].Contact]:
            Self.Canvas.Create_Scaled_Line(X+Last[0], Y+Last[1], X+Point[0], Y+Point[1], fill=Color, tag=ID)
            Last = Point

    def Highlight(Self, On) :
        """ This routine highlights or unhighlights a wire object. """
        if On :
            ID = Self.ID
            Self.HighlightID = []
            Last = Self.Terminals[0].Contact
            for Point in Self.Points + [Self.Terminals[1].Contact]:
                Pm = [0,0]
                Pa = [0,0]
                Pb = [0,0]
                for I in (0, 1) :
                    if Last[I] < Point[I] :
                        Pm[I] = -3
                    else :
                        Pm[I] = 3
                    Pa[I] = Self.Base[I] + Last[I] + Pm[I]
                    Pb[I] = Self.Base[I] + Point[I] - Pm[I]
                HID = Self.Canvas.Create_Scaled_Rectangle(Pa[0], Pa[1], Pb[0], Pb[1], width=0, fill="lightgray", tag=ID)
                Self.Canvas.tag_lower(HID, ID)
                Self.HighlightID.append(HID)
                Last = Point
        else :
            for HID in Self.HighlightID :
                Self.Canvas.delete(HID)
            Self.HighlightID = None

    def Intersect(Self, Point) :
        """ This routine determines whether a point intersects with a drawn segment. """
        LastX, LastY = Self.Terminals[0].Contact
        LastX += Self.Base[0]
        LastY += Self.Base[1]
        for X, Y in Self.Points + [Self.Terminals[1].Contact]:
            X += Self.Base[0]
            Y += Self.Base[1]
            if Self.Canvas.PointOnLine(Point, (LastX,LastY), (X,Y)) :
                return True
            LastX = X
            LastY = Y
        return False

    def Split(Self, CutPoint) :
        """ This routine splits a wire at a specified point on the drawn segments.
        Two wires are returned. The specified points becomes an endpoint terminal
        for each wire. """
        for I in (0, 1) :
            CutPoint[I] -= Self.Base[I]
        Points = copy([Self.Terminals[0].Contact] + Self.Points + [Self.Terminals[1].Contact])  # is copy needed?
        if CutPoint not in Points :
            Last = Points[0]
            for Point in Points :
                if Self.Canvas.PointOnLine(CutPoint, Last, Point) :
                    I = Points.index(Last)
                    Points = Points[:I+1] + [CutPoint] + Points[I+1:]
                    break
                Last = Point
        I = Points.index(CutPoint)
        Wire1 = WIRE(Self.Canvas, Base = Self.Base, Points = Points[:I] + [copy(CutPoint)])
        Wire2 = WIRE(Self.Canvas, Base = copy(Self.Base), Points = Points[I:])
        Wire1.Draw()
        Wire2.Draw()
        return [Wire1, Wire2]

    def Write(Self, File) :
        """ This routine writes a wire object to the output file. """
        File.write("%s(Base=%s, " % (Self.Type, Self.Base))
        File.write("Points= [%s, " % (Self.Terminals[0].Contact))
        for Point in Self.Points :
            File.write("%s, " % (Point))
        File.write("%s])\n" % (Self.Terminals[1].Contact))

    def Clone(Self) :
        """ This routine returns a clone string. """
        String = "%s(Self.Canvas, Base=%s, Points= [%s, " % (Self.Type, Self.Base, Self.Terminals[0].Contact)
        for Point in Self.Points :
            String += "%s, " % (Point)
        String += "%s])" % (Self.Terminals[1].Contact)
        return String

class Gate(Object) :
    """ This is a gate type compucanvas object. """
    def __init__(Self, Canvas, Base=[0,0], NumIn=2, BubbleIn=False, BubbleOut=False) :
        Object.__init__(Self, Canvas, Base)
        Self.NumIn = NumIn
        Self.BubbleIn = BubbleIn
        Self.BubbleOut = BubbleOut

    def Properties(Self, ObjList) :
        """ This routine builds a gate property menu. The PropLocks
        list unchangable properties for a given gate type. """
        Self.Highlight(True)
        PropertiesForm = Toplevel()
        PropertiesForm.title('Properties')

        NumIn = IntVar()
        NumIn.set(Self.NumIn)
        if Self.PropLocks[0] == 0 :
            NumInFrame = Frame(PropertiesForm)
            NumInFrame.pack(side=TOP, fill=X)
            NumInLabel = Label(NumInFrame, text='number of inputs', width=15)
            NumInLabel.pack(side=LEFT)
            for Size in (2, 3) :
                NumInEntry = Radiobutton(NumInFrame, text=str(Size), variable=NumIn, value=Size)
                NumInEntry.pack(side=LEFT)

        BubbleIn = IntVar()
        BubbleIn.set(Self.BubbleIn)
        if Self.PropLocks[1] == 0 :
            BubbleInFrame = Frame(PropertiesForm)
            BubbleInFrame.pack(side=TOP, fill=X)
            BubbleInLabel = Label(BubbleInFrame, text='bubbled inputs?', width=15)
            BubbleInLabel.pack(side=LEFT)
            for Option, Value in (('yes', True), ('no', False)) :
                BubbleInEntry = Radiobutton(BubbleInFrame, text=Option, variable=BubbleIn, value=Value)
                BubbleInEntry.pack(side=LEFT)

        BubbleOut = IntVar()
        BubbleOut.set(Self.BubbleOut)
        if Self.PropLocks[2] == 0 :
            BubbleOutFrame = Frame(PropertiesForm)
            BubbleOutFrame.pack(side=TOP, fill=X)
            BubbleOutLabel = Label(BubbleOutFrame, text='bubbled output?', width=15)
            BubbleOutLabel.pack(side=LEFT)
            for Option, Value in (('yes', True), ('no', False)) :
                BubbleOutEntry = Radiobutton(BubbleOutFrame, text=Option, variable=BubbleOut, value=Value)
                BubbleOutEntry.pack(side=LEFT)

        Parameters = (NumIn, BubbleIn, BubbleOut)
        ButtonFrame = Frame(PropertiesForm)
        ButtonFrame.pack(side=TOP, fill=X)
        OKButton = Button(ButtonFrame, text='OK', command=(lambda Vars=Parameters,
                                                           Win=PropertiesForm,
                                                           List=ObjList:
                                                           Self.Update_Properties(Vars, Win, List)))
        OKButton.pack(side=LEFT, expand=YES, fill=X)
        CancelButton = Button(ButtonFrame, text='Cancel',  command=(lambda Vars=[],
                                                                    Win=PropertiesForm,
                                                                    List=ObjList:
                                                                    Self.Update_Properties(Vars, Win, ObjList)))
        CancelButton.pack(side=RIGHT, expand=YES, fill=X)

    def Update_Properties(Self, Vars, Win, ObjList) :
        """ This routine updates gate properties if a list is provided.
        If new parameters are provided, it compares new versus current
        parameters and updates them if they are different. If the PropLock
        is true, the current value is unchanged. After new parameters are
        processed (or if they are not provided), the object update_properties
        routine is called to implement the object list call chain. """
        if Vars :
            NumIn = Vars[0].get()
            BubbleIn = Vars[1].get()
            BubbleOut = Vars[2].get()
            if Self.PropLocks[0] <> 1 :
                Self.NumIn = NumIn
                Self.Terminals[1].Active = (NumIn == 3)
            if Self.PropLocks[1] <> 1 :
                Self.BubbleIn = BubbleIn
                if len(Self.Terminals) == 4 :
                    for I in range(3) :
                        Self.Terminals[I].Bubble = BubbleIn
                else :
                    Self.Terminals[0].Bubble = BubbleIn
            if Self.PropLocks[2] <> 1 :
                Self.BubbleOut = BubbleOut
                Self.Terminals[-1].Bubble = BubbleOut
        Object.Update_Properties(Self, Vars, Win, ObjList)

    def Write(Self, File) :
        """ This routine writes a gate object to the output file. """
        File.write("%s(Base=%s, NumIn=%s, BubbleIn=%s, BubbleOut=%s)\n" % \
                   (Self.Type, Self.Base, Self.NumIn, Self.BubbleIn, Self.BubbleOut))

    def Clone(Self) :
        """ This routine returns a clone string. """
        return "%s(Self.Canvas, Base=%s, NumIn=%s, BubbleIn=%s, BubbleOut=%s)\n" % \
               (Self.Type, Self.Base, Self.NumIn, Self.BubbleIn, Self.BubbleOut)

class AND(Gate) :
    def __init__(Self, Canvas, Base=[0,0], NumIn=2, BubbleIn=False, BubbleOut=False) :
        Gate.__init__(Self, Canvas, Base, NumIn, BubbleIn, BubbleOut)
        Self.Type = 'AND'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.PropLocks = (0,0,0)
        Self.Size = (90, 40)
        Self.Terminals = (Terminal(Self, [0,10], [20,10], Bubble=BubbleIn),
                          Terminal(Self, [0,20], [20,20], Active=NumIn>2, Bubble=BubbleIn),
                          Terminal(Self, [0,30], [20,30], Bubble=BubbleIn),
                          Terminal(Self, [90,20], [70,20], Bubble=BubbleOut))

    def Simulate(Self) :
        """ This routine simulates an AND gate. Input inversion
        is achived by selecting the match state. Then the AND function
        is computed as an OR predicate since a single input can
        determine the output. Bad inputs (float, short, and unknown)
        are also detected. If the AND function is true (no false inputs)
        and bad inputs are detected, the result will be unknown. Otherwise
        the AND output is either false (because of a false input) or
        all inputs are true (with no bad inputs). In this case, output
        inversation and the resulting state define the result. The result
        then becomes the source for the output terminal. """
        if Self.BubbleIn :
            MatchState = 1
        else :
            MatchState = 0
        State = True
        BadIn = False
        for Terminal in Self.Terminals[:3] :
            if Terminal.Active :
                if Terminal.Node.State == MatchState :
                    State = False
                    break
                elif Terminal.Node.State > 1 :
                    BadIn = True
        if State and BadIn :
            Source = 'X'
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
        else :
            if Self.BubbleOut :
                State = not State
            if State :
                Source = '1'
            else :
                Source = '0'
        Self.Terminals[-1].Source = Source
        return []

    def Draw(Self) :
        """ This routine draws an AND body gate. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Line(X+20, Y, X+20, Y+40, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+20, Y, X+50, Y, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+20, Y+40, X+50, Y+40, tag=ID)
        Self.Canvas.Create_Scaled_Arc(X+30, Y, X+70, Y+40, start=-90, extent=180, style=ARC, tag=ID)
        if Self.BubbleIn :
            Self.Canvas.Create_Scaled_Oval(X+10, Y+5, X+20, Y+15, tag=ID)
            Self.Canvas.Create_Scaled_Oval(X+10, Y+25, X+20, Y+35, tag=ID)
            if Self.NumIn == 3 :
                Self.Canvas.Create_Scaled_Oval(X+10, Y+15, X+20, Y+25, tag=ID)
        if Self.BubbleOut :
            Self.Canvas.Create_Scaled_Oval(X+70, Y+15, X+80, Y+25, tag=ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()

class OR(Gate) :
    def __init__(Self, Canvas, Base=[0,0], NumIn=2, BubbleIn=False, BubbleOut=False) :
        Gate.__init__(Self, Canvas, Base, NumIn, BubbleIn, BubbleOut)
        Self.Type = 'OR'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.PropLocks = (0,0,0)
        Self.Size = (90, 40)
        Self.Terminals = (Terminal(Self, [0,10], [28,10], Bubble=BubbleIn),
                          Terminal(Self, [0,20], [30,20], Active=NumIn>2, Bubble=BubbleIn),
                          Terminal(Self, [0,30], [28,30], Bubble=BubbleIn),
                          Terminal(Self, [90,20], [70,20], Bubble=BubbleOut))

    def Simulate(Self) :
        """ This routine simulates an OR gate. Input inversion
        is achived by selecting the match state. Then the OR function
        is computed. Bad inputs (float, short, and unknown)
        are also detected. If the OR function is false (no true inputs)
        and bad inputs are detected, the result will be unknown. Otherwise
        the OR output is either true (because of a true input) or
        all inputs are false (with no bad inputs). In this case, output
        inversation and the resulting state define the result. The result
        then becomes the source for the output terminal. """
        if Self.BubbleIn :
            MatchState = 0
        else :
            MatchState = 1
        State = False
        BadIn = False
        for Terminal in Self.Terminals[:3] :
            if Terminal.Active :
                if Terminal.Node.State == MatchState :
                    State = True
                    break
                elif Terminal.Node.State > 1 :
                    BadIn = True
        if not State and BadIn :
            Source = 'X'
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
        else :
            if Self.BubbleOut :
                State = not State
            if State :
                Source = '1'
            else :
                Source = '0'
        Self.Terminals[-1].Source = Source
        return []

    def Draw(Self) :
        """ This routine draws an OR body gate. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Arc(X-30, Y, X+70, Y+40, start=-90, extent=180, style=ARC, tag=ID)
        Self.Canvas.Create_Scaled_Arc(X+10, Y, X+30, Y+40, start=-90, extent=180, style=ARC, tag=ID)
        if Self.BubbleIn :
            Self.Canvas.Create_Scaled_Oval(X+18, Y+5, X+28, Y+15, tag=ID)
            Self.Canvas.Create_Scaled_Oval(X+18, Y+25, X+28, Y+35, tag=ID)
            if Self.NumIn == 3 :
                Self.Canvas.Create_Scaled_Oval(X+20, Y+15, X+30, Y+25, tag=ID)
        if Self.BubbleOut :
            Self.Canvas.Create_Scaled_Oval(X+70, Y+15, X+80, Y+25, tag=ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()

class XOR(Gate) :
    def __init__(Self, Canvas, Base=[0,0], NumIn=2, BubbleIn=False, BubbleOut=False) :
        Gate.__init__(Self, Canvas, Base, NumIn, BubbleIn, BubbleOut)
        Self.Type = 'XOR'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.PropLocks = (0,1,0)
        Self.Size = (90, 40)
        Self.Terminals = (Terminal(Self, [0,10], [18,10]),
                          Terminal(Self, [0,20], [20,20], Active=NumIn>2),
                          Terminal(Self, [0,30], [18,30]),
                          Terminal(Self, [90,20], [70,20], Bubble=BubbleOut))

    def Simulate(Self) :
        """ This routine simulates an XOR gate. Then the OR function
        is computed. Bad inputs (float, short, and unknown) are also
        detected. True inputs toggle the state. A bad input results in
        an unknown output. Otherwise the state and output inversion define
        the result. The result then becomes the source for the output terminal. """
        State = False
        BadIn = False
        for Terminal in Self.Terminals[:3] :
            if Terminal.Active :
                if Terminal.Node.State == 1 :
                    State = not State
                if Terminal.Node.State > 1 :
                    BadIn = True
                    break
        if BadIn :
            Source = 'X'
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
        else :
            if Self.BubbleOut :
                State = not State
            if State :
                Source = '1'
            else :
                Source = '0'
        Self.Terminals[-1].Source = Source
        return []

    def Draw(Self) :
        """ This routine draws an XOR body gate. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Arc(X-30, Y, X+70, Y+40, start=-90, extent=180, style=ARC, tag=ID)
        Self.Canvas.Create_Scaled_Arc(X+10, Y, X+30, Y+40, start=-90, extent=180, style=ARC, tag=ID)
        Self.Canvas.Create_Scaled_Arc(X, Y, X+20, Y+40, start=-90, extent=180, style=ARC, tag=ID)
        if Self.BubbleOut :
            Self.Canvas.Create_Scaled_Oval(X+70, Y+15, X+80, Y+25, tag=ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()

class NOT(Gate) :
    def __init__(Self, Canvas, Base=[0,0], NumIn=1, BubbleIn=False, BubbleOut=False) :
        Gate.__init__(Self, Canvas, Base, NumIn, BubbleIn, BubbleOut)
        Self.Type = 'NOT'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.PropLocks = (1,0,0)
        Self.Size = (60, 20)
        Self.Terminals = (Terminal(Self, [0,10], [20,10], Bubble=BubbleIn),
                          Terminal(Self, [60,10], [40,10], Bubble=BubbleOut))

    def Simulate(Self) :
        """ This routine simulates a NOT gate. A bad input (float, short,
        and unknown) produces an unknown result. Otherwise the input is
        inverted. Input and output inversions also toggle the state. The
        result then becomes the source for the output terminal. """
        Terminal = Self.Terminals[0]
        State = False
        if Terminal.Node.State > 1 :
            Source = 'X'
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
        else :
            if Terminal.Node.State :
                State = True
            if Self.BubbleIn :
                State = not State
            if Self.BubbleOut :
                State = not State
            if State :
                Source = '1'
            else :
                Source = '0'
        Self.Terminals[-1].Source = Source
        return []

    def Draw(Self) :
        """ This routine draws a NOT body gate. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Line(X+20, Y, X+20, Y+20, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+20, Y, X+40, Y+10, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+20, Y+20, X+40, Y+10, tag=ID)
        if Self.BubbleIn :
            Self.Canvas.Create_Scaled_Oval(X+10, Y+5, X+20, Y+15, tag=ID)
        if Self.BubbleOut :
            Self.Canvas.Create_Scaled_Oval(X+40, Y+5, X+50, Y+15, tag=ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()

class BAR(Gate) :
    def __init__(Self, Canvas, Base=[0,0], NumIn=1, BubbleIn=False, BubbleOut=False) :
        Gate.__init__(Self, Canvas, Base, NumIn, BubbleIn, BubbleOut)
        Self.Type = 'BAR'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.PropLocks = (1,0,0)
        Self.Size = (40, 20)
        Self.Terminals = (Terminal(Self, [0,10], [20,10], Bubble=BubbleIn),
                          Terminal(Self, [40,10], [20,10], Bubble=BubbleOut))

    def Draw(Self) :
        """ This routine draws a bubbled BAR. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Rectangle(X+20, Y, X+20, Y+20, tag=ID)
        if Self.BubbleIn :
            Self.Canvas.Create_Scaled_Oval(X+10, Y+5, X+20, Y+15, tag=ID)
        if Self.BubbleOut :
            Self.Canvas.Create_Scaled_Oval(X+20, Y+5, X+30, Y+15, tag=ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()

class FET(Gate) :
    def __init__(Self, Canvas, Base=[0,0], NumIn=1, BubbleIn=False, BubbleOut=False) :
        Gate.__init__(Self, Canvas, Base, NumIn, BubbleIn, BubbleOut)
        Self.Type = 'FET'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.PropLocks = (1,0,1)
        Self.Size = (60, 60)
        Self.Switch = False
        Self.Terminals = (Terminal(Self, [0,30], [20,30], Bubble=BubbleIn),
                          Terminal(Self, [60,0], [60,10]),
                          Terminal(Self, [60,60], [60,50]))

    def Simulate(Self) :
        """ This routine simulates a FET. A bad input (float, short,
        and unknown) produces an open switch. Otherwise the input is
        adjusted for input inversion and then used to control switch
        state. A Cnode list is returned if the switch is closed. """
        Terminal = Self.Terminals[0]
        Self.Switch = False
        if Terminal.Node.State > 1 :
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
            return []
        else :
            if Terminal.Node.State :
                Self.Switch = True
            if Self.BubbleIn :
                Self.Switch = not Self.Switch
            if Self.Switch :
                return [Self.Terminals[1].Node, Self.Terminals[2].Node]
            else :
                return []

    def Draw(Self) :
        """ This routine draws an FET. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Line(X+20, Y+10, X+20, Y+50, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+30, Y+10, X+30, Y+50, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+30, Y+10, X+60, Y+10, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+30, Y+50, X+60, Y+50, tag=ID)
        if Self.BubbleIn :
            Self.Canvas.Create_Scaled_Oval(X+10, Y+25, X+20, Y+35, tag=ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()
        if Self.Canvas.master.SwitchMarker and Self.Switch :
            Self.Canvas.Create_Scaled_Line(X+60, Y+13, X+60, Y+47, fill="cyan", tag=ID)

class DECODER(Object) :
    def __init__(Self, Canvas, Base=[0,0], NumIn=1) :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'DECODER'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (60, 80)
        Self.NumIn = NumIn
        Self.Terminals = (Terminal(Self, [0,20], [10,20]),
                          Terminal(Self, [0,40], [10,40], Active = Self.NumIn > 1),
                          Terminal(Self, [0,60], [10,60]),
                          Terminal(Self, [60,10], [50,10], Active = Self.NumIn > 1),
                          Terminal(Self, [60,30], [50,30]),
                          Terminal(Self, [60,50], [50,50]),
                          Terminal(Self, [60,70], [50,70], Active = Self.NumIn > 1))

    def Simulate(Self) :
        """ This routine simulates a decoder. If the enable input is high,
        the output corrsponding to the input value is set high. Otherwise
        all outputs are low. The number of inputs can be one or two with
        two or four outputs. """

        if  Self.NumIn == 1 :
            if Self.Terminals[0].Node.State > 1 or Self.Terminals[2].Node.State > 1 :
                Self.Terminals[4].Source = Self.Terminals[5].Source = 'X'
                Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
            else :
                Self.Terminals[4].Source = '0'
                Self.Terminals[5].Source = '0'
                if Self.Terminals[2].Node.State == 1 :
                    if Self.Terminals[0].Node.State == 1 :
                        Self.Terminals[5].Source = '1'
                    else :
                        Self.Terminals[4].Source = '1'
        if Self.NumIn == 2 :
            if Self.Terminals[0].Node.State > 1 or Self.Terminals[1].Node.State > 1 or Self.Terminals[2].Node.State > 1 :
                Self.Terminals[3].Source = Self.Terminals[4].Source = Self.Terminals[5].Source = Self.Terminals[6].Source = 'X'
                Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
                return
            else :
                Self.Terminals[3].Source = '0'
                Self.Terminals[4].Source = '0'
                Self.Terminals[5].Source = '0'
                Self.Terminals[6].Source = '0'
                Out = 0
                if Self.Terminals[2].Node.State == 1 :
                    if Self.Terminals[0].Node.State == 1 :
                        Out += 1
                    if Self.Terminals[1].Node.State == 1 :
                        Out += 2
                    Self.Terminals[Out + 3].Source = '1'
        return []

    def Properties(Self, ObjList) :
        """ This routine builds a decoder property menu. """
        Self.Highlight(True)
        PropertiesForm = Toplevel()
        PropertiesForm.title('Properties')

        NumIn = IntVar()
        NumIn.set(Self.NumIn)
        NumInFrame = Frame(PropertiesForm)
        NumInFrame.pack(side=TOP, fill=X)
        NumInLabel = Label(NumInFrame, text='number of inputs', width=15)
        NumInLabel.pack(side=LEFT)
        for Size in (1, 2) :
            NumInEntry = Radiobutton(NumInFrame, text=str(Size), variable=NumIn, value=Size)
            NumInEntry.pack(side=LEFT)

        ButtonFrame = Frame(PropertiesForm)
        ButtonFrame.pack(side=TOP, fill=X)

        OKButton = Button(ButtonFrame, text='OK',  command=lambda: Self.Set_Decoder_Inputs(NumIn, PropertiesForm, ObjList))
        OKButton.pack(side=LEFT, expand=YES, fill=X)
        CancelButton = Button(ButtonFrame, text='Cancel', command=lambda: Self.Set_Decoder_Inputs([], PropertiesForm, ObjList))
        CancelButton.pack(side=RIGHT, expand=YES, fill=X)

    def Set_Decoder_Inputs(Self, NumIn, Win, ObjList) :
        """ This routine updates the number of inputs property.
        After processing (or if they are not provided), the
        object update_properties routine is called to implement
        the object list call chain. """
        if NumIn <> [] :
            Self.NumIn = NumIn.get()
            Self.Terminals[1].Active = (Self.NumIn == 2)
            Self.Terminals[3].Active = (Self.NumIn == 2)
            Self.Terminals[6].Active = (Self.NumIn == 2)
        Object.Update_Properties(Self, NumIn, Win, ObjList)

    def Draw(Self) :
        """ This routine draws a decoder. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+50, Y+00, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+80, X+50, Y+80, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+10, Y+80, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+50, Y+00, X+50, Y+80, tag=ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+20, text='I0', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        if Self.NumIn > 1 :
            Self.Canvas.Create_Scaled_Text(X+13, Y+40, text='I1', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+60, text='EN', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        if Self.NumIn > 1 :
            Self.Canvas.Create_Scaled_Text(X+47, Y+10, text='O0', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+47, Y+30, text='O1', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+47, Y+50, text='O2', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+47, Y+70, text='O3', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
        else :
            Self.Canvas.Create_Scaled_Text(X+47, Y+30, text='O0', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+47, Y+50, text='O1', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+35, Y+40, text='Dec', anchor='c', font=Self.Canvas.Bold, tag=Self.ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()

    def Write(Self, File) :
        """ This routine writes a deocder object to the output file. """
        File.write("%s(Base=%s, NumIn=%s)\n" % (Self.Type, Self.Base, Self.NumIn))

    def Clone(Self) :
        """ This routine returns a clone string. """
        return "%s(Self.Canvas, Base=%s, NumIn=%s)\n" % (Self.Type, Self.Base, Self.NumIn)

class MUX(Object) :
    def __init__(Self, Canvas, Base=[0,0], NumIn=2) :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'MUX'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (60, 50)
        Self.NumIn = NumIn
        Self.Terminals = (Terminal(Self, [20,90], [20,80], Active = Self.NumIn == 4),
                          Terminal(Self, [30,50], [30,40], Active = Self.NumIn == 2),
                          Terminal(Self, [40,90], [40,80], Active = Self.NumIn == 4),
                          Terminal(Self, [0,10], [10,10]),
                          Terminal(Self, [0,30], [10,30]),
                          Terminal(Self, [0,50], [10,50], Active = Self.NumIn == 4),
                          Terminal(Self, [0,70], [10,70], Active = Self.NumIn == 4),
                          Terminal(Self, [60,20], [50,20], Active = Self.NumIn == 2),
                          Terminal(Self, [60,40], [50,40], Active = Self.NumIn == 4))

    def Simulate(Self) :
        """ This routine simulates a mux/demux. It can have either two or
        four inputs/outputs. It handles invalid inputs, then the output
        corrsponding to the select value is connected to the input using
        cnodes. """

        if  Self.NumIn == 2 :
            if Self.Terminals[1].Node.State > 1 :
                Self.Terminals[7].Source = 'X'
                return []
            elif Self.Terminals[1].Node.State == 0 :
                return [Self.Terminals[7].Node, Self.Terminals[3].Node]
            elif Self.Terminals[1].Node.State == 1 :
                return [Self.Terminals[7].Node, Self.Terminals[4].Node]
        if Self.NumIn == 4 :
            if Self.Terminals[0].Node.State > 1 or Self.Terminals[2].Node.State > 1 :
                Self.Terminals[7].Source = 'X'
                return []
            elif Self.Terminals[0].Node.State == 0 and Self.Terminals[2].Node.State == 0 :
                return [Self.Terminals[8].Node, Self.Terminals[3].Node]
            elif Self.Terminals[0].Node.State == 1 and Self.Terminals[2].Node.State == 0 :
                return [Self.Terminals[8].Node, Self.Terminals[4].Node]
            elif Self.Terminals[0].Node.State == 0 and Self.Terminals[2].Node.State == 1 :
                return [Self.Terminals[8].Node, Self.Terminals[5].Node]
            elif Self.Terminals[0].Node.State == 1 and Self.Terminals[2].Node.State == 1 :
                return [Self.Terminals[8].Node, Self.Terminals[6].Node]

    def Properties(Self, ObjList) :
        """ This routine builds a mux property menu. """
        Self.Highlight(True)
        PropertiesForm = Toplevel()
        PropertiesForm.title('Properties')
        NumIn = IntVar()
        NumIn.set(Self.NumIn)
        NumInFrame = Frame(PropertiesForm)
        NumInFrame.pack(side=TOP, fill=X)
        NumInLabel = Label(NumInFrame, text='number of inputs', width=15)
        NumInLabel.pack(side=LEFT)
        for Size in (2, 4) :
            NumInEntry = Radiobutton(NumInFrame, text=str(Size), variable=NumIn, value=Size)
            NumInEntry.pack(side=LEFT)
        ButtonFrame = Frame(PropertiesForm)
        ButtonFrame.pack(side=TOP, fill=X)
        OKButton = Button(ButtonFrame, text='OK',  command=lambda: Self.Set_Mux_Inputs(NumIn, PropertiesForm, ObjList))
        OKButton.pack(side=LEFT, expand=YES, fill=X)
        CancelButton = Button(ButtonFrame, text='Cancel', command=lambda: Self.Set_Mux_Inputs([], PropertiesForm, ObjList))
        CancelButton.pack(side=RIGHT, expand=YES, fill=X)

    def Set_Mux_Inputs(Self, NumIn, Win, ObjList) :
        """ This routine updates the number of inputs property.
        After processing (or if they are not provided), the
        object update_properties routine is called to implement
        the object list call chain. """
        if NumIn <> [] :
            Self.NumIn = NumIn.get()
            Self.Terminals[0].Active = (Self.NumIn == 4)
            Self.Terminals[1].Active = (Self.NumIn == 2)
            Self.Terminals[2].Active = (Self.NumIn == 4)
            Self.Terminals[5].Active = (Self.NumIn == 4)
            Self.Terminals[6].Active = (Self.NumIn == 4)
            Self.Terminals[7].Active = (Self.NumIn == 2)
            Self.Terminals[8].Active = (Self.NumIn == 4)
        Object.Update_Properties(Self, NumIn, Win, ObjList)

    def Draw(Self) :
        """ This routine draws a mux. """
        X, Y = Self.Base
        ID = Self.ID
        if Self.NumIn == 4 :
            Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+50, Y+00, tag=ID)
            Self.Canvas.Create_Scaled_Line(X+10, Y+80, X+50, Y+80, tag=ID)
            Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+10, Y+80, tag=ID)
            Self.Canvas.Create_Scaled_Line(X+50, Y+00, X+50, Y+80, tag=ID)
            Self.Canvas.Create_Scaled_Text(X+13, Y+10, text='I0', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+13, Y+30, text='I1', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+13, Y+50, text='I2', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+13, Y+70, text='I3', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+20, Y+80, text='S0', anchor='s', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+40, Y+80, text='S1', anchor='s', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+47, Y+40, text='Out', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+30, Y+20, text='Mux', anchor='c', font=Self.Canvas.Bold, tag=Self.ID)
            Self.Size = (60, 90)
        elif Self.NumIn == 2 :
            Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+50, Y+00, tag=ID)
            Self.Canvas.Create_Scaled_Line(X+10, Y+40, X+50, Y+40, tag=ID)
            Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+10, Y+40, tag=ID)
            Self.Canvas.Create_Scaled_Line(X+50, Y+00, X+50, Y+40, tag=ID)
            Self.Canvas.Create_Scaled_Text(X+13, Y+10, text='I0', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+13, Y+30, text='I1', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+30, Y+40, text='S0', anchor='s', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+47, Y+20, text='Out', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+20, Y+20, text='Mux', anchor='c', font=Self.Canvas.Bold, tag=Self.ID)
            Self.Size = (60, 50)
        for Terminal in Self.Terminals :
            Terminal.Draw()

    def Write(Self, File) :
        """ This routine writes a mux object to the output file. """
        File.write("%s(Base=%s, NumIn=%s)\n" % (Self.Type, Self.Base, Self.NumIn))

    def Clone(Self) :
        """ This routine returns a clone string. """
        return "%s(Self.Canvas, Base=%s, NumIn=%s)\n" % (Self.Type, Self.Base, Self.NumIn)

class ARITH(Object) :
    def __init__(Self, Canvas, Base=[0,0], Func="ADD") :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'ARITH'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (60, 60)
        Self.Func = Func
        Self.Terminals = (Terminal(Self, [0,20], [10,20]),  # 0: X input
                          Terminal(Self, [0,40], [10,40]),  # 1: Y input
                          Terminal(Self, [30,0], [30,10]),  # 2: Cin/Bin
                          Terminal(Self, [60,30], [50,30]), # 3: Sum/Dif
                          Terminal(Self, [30,60], [30,50])) # 4: Cout/Bout

    def Simulate(Self) :
        """ This routine simulates an arithmatic unit (adder or subtractor).
        For "ADD":
            if X, Y, and Cin are valid, S = xor
            else S = invalid
            if two of X, Y, and Cin are high, Cout = 1
            elif X, Y, and Cin are valid, Cout = 0
            elif Cout = invalid
        For "SUB":
            if X, Y, and Bin are valid, D = xor
            else D = invalid
            if -XY, -XBin, or YBin are high, Bout = 1
            elif X, Y, and Cin are valid, Cout = 0
            elif Cout = invalid """

        BadInput = False
        if Self.Terminals[0].Node.State > 1 or Self.Terminals[1].Node.State > 1 or \
           Self.Terminals[2].Node.State > 1 :
            BadInput = True
        OneCount = 0
        if Self.Terminals[0].Node.State == 1 :
            OneCount += 1
        if Self.Terminals[1].Node.State == 1 :
            OneCount += 1
        if Self.Terminals[2].Node.State == 1 :
            OneCount += 1
        if BadInput :
            Self.Terminals[3].Source = 'X'
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
        else :
            Self.Terminals[3].Source = str(OneCount % 2)
        if  Self.Func == "ADD" :
            if OneCount >= 2 :
                Self.Terminals[4].Source = '1'
            elif BadInput :
                Self.Terminals[4].Source = 'X'
                Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
            else :
                Self.Terminals[4].Source = '0'
        if  Self.Func == "SUB" :
            if Self.Terminals[1].Node.State == 1 and Self.Terminals[2].Node.State == 1 or \
               Self.Terminals[0].Node.State == 0 and \
               (Self.Terminals[1].Node.State == 1 or Self.Terminals[2].Node.State == 1) :
                Self.Terminals[4].Source = '1'
            elif BadInput :
                Self.Terminals[4].Source = 'X'
                Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
            else :
                Self.Terminals[4].Source = '0'
        return []

    def Properties(Self, ObjList) :
        """ This routine builds an arithmatic unit property menu. """
        Self.Highlight(True)
        PropertiesForm = Toplevel()
        PropertiesForm.title('Properties')

        Func = StringVar()
        Func.set(Self.Func)
        FuncFrame = Frame(PropertiesForm)
        FuncFrame.pack(side=TOP, fill=X)
        FuncLabel = Label(FuncFrame, text='position', width=15)
        FuncLabel.pack(side=LEFT)
        for Pos in ('ADD', 'SUB') :
            FuncEntry = Radiobutton(FuncFrame, text=Pos, variable=Func, value=Pos)
            FuncEntry.pack(side=LEFT)

        ButtonFrame = Frame(PropertiesForm)
        ButtonFrame.pack(side=TOP, fill=X)
        OKButton = Button(ButtonFrame, text='OK',  command=lambda: Self.Set_Arith_Func(Func, PropertiesForm, ObjList))
        OKButton.pack(side=LEFT, expand=YES, fill=X)
        CancelButton = Button(ButtonFrame, text='Cancel', command=lambda: Self.Set_Arith_Func([], PropertiesForm, ObjList))
        CancelButton.pack(side=RIGHT, expand=YES, fill=X)

    def Set_Arith_Func(Self, Func, Win, ObjList) :
        """ This routine updates the function property "ADD" or
        "SUB". After processing (or if they are not provided),
        the object update_properties routine is called to implement
        the object list call chain. """
        if Func <> [] :
            Self.Func = Func.get()
        Object.Update_Properties(Self, Func, Win, ObjList)

    def Draw(Self) :
        """ This routine draws an arithmetic unit. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Line(X+10, Y+10, X+50, Y+10, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+50, X+50, Y+50, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+10, X+10, Y+50, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+50, Y+10, X+50, Y+50, tag=ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+20, text='X', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+40, text='Y', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        if Self.Func == "ADD" :
            Self.Canvas.Create_Scaled_Text(X+47, Y+30, text='S', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+30, Y+50, text='Cout', anchor='s', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+30, Y+10, text='Cin', anchor='n', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+30, Y+30, text='+', anchor='c', font=Self.Canvas.Bold, tag=Self.ID)
        elif Self.Func == "SUB" :
            Self.Canvas.Create_Scaled_Text(X+47, Y+30, text='D', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+30, Y+50, text='Bout', anchor='s', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+30, Y+10, text='Bin', anchor='n', font=Self.Canvas.Font, tag=Self.ID)
            Self.Canvas.Create_Scaled_Text(X+30, Y+30, text='-', anchor='c', font=Self.Canvas.Bold, tag=Self.ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()

    def Write(Self, File) :
        """ This routine writes an arithmetic unit to the output file. """
        File.write("%s(Base=%s, Func= '%s')\n" % (Self.Type, Self.Base, Self.Func))

    def Clone(Self) :
        """ This routine returns a clone string. """
        return "%s(Self.Canvas, Base=%s, Func= '%s')\n" % (Self.Type, Self.Base, Self.Func)

class LATCH(Object) :
    def __init__(Self, Canvas, Base=[0,0]) :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'LATCH'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (60, 80)
        Self.State = 'X'
        Self.Terminals = (Terminal(Self, [0,20], [10,20]),  # 0: IN
                          Terminal(Self, [30,80], [30,70]), # 1: EN
                          Terminal(Self, [60,20], [50,20])) # 2: OUT

    def Simulate(Self) :
        """ This routine simulates a transparent latch. If the enable
        input is high, the output follows the input. Otherwise when
        enable is low, the output retains the previous value. When no
        value has been set, the value is unknown: 'X'. """   
        if Self.Terminals[1].Node.State > 1 :
            Self.Terminals[2].Source = 'X'
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
        elif Self.Terminals[1].Node.State == 1 :
            if Self.Terminals[0].Node.State > 1 :
                Self.Terminals[2].Source = 'X'
                Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
            elif Self.Terminals[0].Node.State == 1 :
                Self.State = Self.Terminals[2].Source = '1'
            else :
                Self.State = Self.Terminals[2].Source = '0'
        elif Self.Terminals[1].Node.State == 0 :
            Self.Terminals[2].Source = Self.State
        return []

    def Draw(Self) :
        """ This routine draws a LATCH. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+50, Y+00, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+70, X+50, Y+70, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+10, Y+70, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+50, Y+00, X+50, Y+70, tag=ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+20, text='I', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+47, Y+20, text='O', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+30, Y+70, text='E', anchor='s', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+30, Y+35, text='Latch', anchor='c', font=Self.Canvas.Bold, tag=Self.ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()

class REGISTER(Object) :
    def __init__(Self, Canvas, Base=[0,0]) :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'REGISTER'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (60, 80)
        Self.InState = 'X'
        Self.OutState = 'X'
        Self.Terminals = (Terminal(Self, [0,20], [10,20]),  # 0: IN
                          Terminal(Self, [0,50], [10,50]),  # 1: WE
                          Terminal(Self, [20,80], [20,70]), # 2: PHI1
                          Terminal(Self, [40,80], [40,70]), # 3: PHI2
                          Terminal(Self, [60,20], [50,20]), # 4: OUT
                          Terminal(Self, [60,50], [50,50])) # 5: RE
        
    def Simulate(Self) :
        """ This routine simulates a regsiter. The register employs
        write enable (WE) to alter the internal state. Phi1 enables
        changes new state. Phi2 enables update output state. When no
        value has been set, the value is unknown: 'X'."""
        if Self.Terminals[3].Node.State > 1 :               # if phi2 == bad state
            Self.Terminals[4].Source = Self.OutState = 'X'  # output unknown
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
        elif Self.Terminals[3].Node.State == 1 :            # phi2 == 1
            if Self.Terminals[5].Node.State > 1 :           # RE == bad state
                Self.Terminals[4].Source = Self.OutState = 'X' # output unknown
                Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
            elif Self.Terminals[5].Node.State == 1 :        # RE == 1
                Self.Terminals[4].Source = Self.OutState = Self.InState # internal state emerges
            elif Self.Terminals[5].Node.State == 0 :        # RE == 0                                    
                Self.Terminals[4].Source = 3                # out floats

        # output of register resolved; now process input

        if Self.Terminals[2].Node.State > 1 :               # if phi1 = bad state
            Self.InState = 'X'                              # set InState = 'X'
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
        elif Self.Terminals[2].Node.State == 1 :            # phi1 == 1
            if Self.Terminals[0].Node.State > 1 and Self.Terminals[1].Node.State == 1 : # In == bad state, WE == 1
                Self.InState = 'X'                          # set InState = 'X'
                Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
            if Self.Terminals[1].Node.State > 1 :           # WE == bad state
                Self.InState = 'X'                          # set InState = 'X'
                Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
            elif Self.Terminals[1].Node.State == 1 :        # WE == 1
                if Self.Terminals[0].Node.State == 1 :       # InState = IN
                    Self.InState = '1'
                elif Self.Terminals[0].Node.State == 0 :
                    Self.InState = '0'
        return []

    def Draw(Self) :
        """ This routine draws a REGISTER. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+50, Y+00, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+70, X+50, Y+70, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+10, Y+70, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+50, Y+00, X+50, Y+70, tag=ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+20, text='In', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+50, text='WE', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+20, Y+70, text='1', anchor='s', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+40, Y+70, text='2', anchor='s', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+47, Y+20, text='Out', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+47, Y+50, text='RE', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+30, Y+35, text='Reg', anchor='c', font=Self.Canvas.Bold, tag=Self.ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()


class TOGGLE(Object) :
    def __init__(Self, Canvas, Base=[0,0]) :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'TOGGLE'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (60, 80)
        Self.InState = 'X'
        Self.OutState = 'X'
        Self.Terminals = (Terminal(Self, [0,20], [10,20]),  # 0: TE
                          Terminal(Self, [0,50], [10,50]),  # 1: -Clr
                          Terminal(Self, [20,80], [20,70]), # 2: Phi1
                          Terminal(Self, [40,80], [40,70]), # 3: Phi2
                          Terminal(Self, [60,20], [50,20])) # 4: Out

    def Simulate(Self) :
        """ This routine simulates a toggle cell. The toggle enable
        (TE) and clear (-Clr) toggle or clear the internal state.
        Clear has higher preceindence. Phi1 enables changes new
        state. Phi2 enables update out output state. When no value has
        been set, the value is unknown: 'X'."""
        if Self.Terminals[3].Node.State > 1 :           # if phi2 == bad state
            Self.Terminals[4].Source = Self.OutState = 'X' # output state and output unknown
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
        elif Self.Terminals[3].Node.State == 1 :        # phi2 == 1
            Self.Terminals[4].Source = Self.OutState = Self.InState  # internal state emerges

        # output of toggle cell resolved; now process input

        if Self.Terminals[2].Node.State > 1 :                # if phi1 = bad state
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
            Self.InState = 'X'                              # set InState = 'X'
        elif Self.Terminals[2].Node.State == 1 :            # phi1 == 1
            if Self.Terminals[1].Node.State > 1 :           # -clr == bad state
                Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
                Self.InState = 'X'                           # set InState = 'X'
            elif Self.Terminals[1].Node.State == 0 :        # -clr == 0 (clear)
                Self.InState = '0'                          # clear internal state
            elif Self.Terminals[1].Node.State == 1 :        # -clr == 1 (pass)
                if Self.Terminals[0].Node.State > 1 :       # toggle enable == bad state
                    Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
                    Self.InState = 'X'                      # set InState = 'X'
                elif Self.Terminals[0].Node.State == 1 :    # toggle enable == 1
                    if Self.OutState == '1':                # toggle output state
                        Self.InState = '0'
                    else :
                        Self.InState = '1'
                else :
                    Self.InState = Self.OutState            # hold output state
        return []

    def Draw(Self) :
        """ This routine draws a TOGGLE. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+50, Y+00, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+70, X+50, Y+70, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+10, Y+70, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+50, Y+00, X+50, Y+70, tag=ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+20, text='TE', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+50, text='Clr', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        H = 4 + Self.Canvas.Font.metrics("ascent") / Self.Canvas.master.Scale
        W = Self.Canvas.Font.measure('Clr') / Self.Canvas.master.Scale
        Self.Canvas.Create_Scaled_Line(X+13, Y+50-H/2, X+13+W, Y+50-H/2, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+20, Y+70, text='1', anchor='s', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+40, Y+70, text='2', anchor='s', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+47, Y+20, text='O', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+30, Y+35, text='Toggle', anchor='c', font=Self.Canvas.Bold, tag=Self.ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()

class COUNTER(Object) :
    def __init__(Self, Canvas, Base=[0,0], N=10) :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'COUNTER'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (60, 100)
        Self.N = N
        Self.InCount = Self.OutCount = 0
        Self.Terminals = (Terminal(Self, [0,40], [10,40]),                          # 0: CE
                          Terminal(Self, [0,70], [10,70]),                          # 1: CLR
                          Terminal(Self, [20,100], [20,90]),                        # 2: Phi1
                          Terminal(Self, [40,100], [40,90]),                        # 3: Phi2
                          Terminal(Self, [60,20], [50,20]),                         # 4: O0
                          Terminal(Self, [60,40], [50,40], Active = Self.N > 2),    # 5: O1
                          Terminal(Self, [60,60], [50,60], Active = Self.N > 4),    # 6: O2
                          Terminal(Self, [60,80], [50,80], Active = Self.N > 8),    # 7: O3
                          Terminal(Self, [30,0], [30,10]))                          # 8: Max Count

    def Simulate(Self) :
        """ This routine simulates a divide by N counter. If there are bad inputs,
        thing don't change (unlcess CLT is unknown). Otherwise during Phi1,
        claer or counting occurs. During Phi2, the new value is output. """
        if Self.Terminals[0].Node.State > 1 or Self.Terminals[2].Node.State > 1 \
           or Self.Terminals[3].Node.State > 1 :                                    # bad clock or CE?
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)             # bad clear?
        elif Self.Terminals[1].Node.State > 1 :
            Self.InCount = 0                                                        # clear count
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
        elif Self.Terminals[2].Node.State == 1 :                                    # phi1?
            if Self.Terminals[1].Node.State == 1 :                                  # and CLR?
                Self.InCount = 0                                                    # clear count
            elif Self.Terminals[0].Node.State == 1 and Self.InCount == Self.OutCount : # else CE and InCount not already incremented?
                Self.InCount = (Self.InCount + 1) % Self.N                          # increment count
        elif Self.Terminals[3].Node.State == 1 :                                    # phi2?
            Self.OutCount = Self.InCount                                            # new count on output
            for I in xrange(4) :                                                    # cool way to define outputs
                Self.Terminals[I + 4].Source = '1' if Self.OutCount & pow(2, I) else '0'
            Self.Terminals[8].Source = '1' if Self.OutCount == Self.N - 1 else '0'  # set max count when appropriate

    def Properties(Self, ObjList) :
        """ This routine builds a decoder property menu. """
        Self.Highlight(True)
        PropertiesForm = Toplevel()
        PropertiesForm.title('Properties')
        Nvar = StringVar()
        Nvar.set(str(Self.N))
        NFrame = Frame(PropertiesForm)
        NFrame.pack(side=TOP, fill=X)
        NLabel = Label(NFrame, text='divide by', width=15)
        NLabel.pack(side=LEFT)
        ButtonFrame = Frame(PropertiesForm)
        ButtonFrame.pack(side=TOP, fill=X)
        NEntry = Entry(NFrame)
        NEntry.config(textvariable=Nvar)
        NEntry.pack(side=RIGHT,  expand=YES, fill=X)
        OKButton = Button(ButtonFrame, text='OK',  command=lambda: Self.Set_N(Nvar, PropertiesForm, ObjList))
        OKButton.pack(side=LEFT, expand=YES, fill=X)
        CancelButton = Button(ButtonFrame, text='Cancel', command=lambda: Self.Set_N([], PropertiesForm, ObjList))
        CancelButton.pack(side=RIGHT, expand=YES, fill=X)

    def Set_N(Self, Nvar, Win, ObjList) :
        """ This routine updates the divide coefficient. """
        if Nvar <> [] :
            try:
                NewN = int(Nvar.get())
                if 2 <= NewN <= 16 :
                    Self.N = NewN
                    Self.Terminals[5].Active = Self.N > 2
                    Self.Terminals[6].Active = Self.N > 4
                    Self.Terminals[7].Active = Self.N > 8
            except ValueError :
                pass
        Object.Update_Properties(Self, Nvar, Win, ObjList)

    def Draw(Self) :
        """ This routine draws a decoder. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Line(X+10, Y+10, X+50, Y+10, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+90, X+50, Y+90, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+10, X+10, Y+90, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+50, Y+10, X+50, Y+90, tag=ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+40, text='CE', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+70, text='CLR', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+20, Y+90, text='1', anchor='s', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+40, Y+90, text='2', anchor='s', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+47, Y+20, text='O0', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+47, Y+40, text='O1', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+47, Y+60, text='O2', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+47, Y+80, text='O3', anchor='e', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+30, Y+10, text='Max', anchor='n', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+30, Y+50, text='/ %d' % (Self.N), anchor='c', font=Self.Canvas.Bold, tag=Self.ID)
        for Terminal in Self.Terminals :
            Terminal.Draw()

    def Write(Self, File) :
        """ This routine writes a counter object to the output file. """
        File.write("%s(Base=%s, N=%s)\n" % (Self.Type, Self.Base, Self.N))

    def Clone(Self) :
        """ This routine returns a clone string. """
        return "%s(Self.Canvas, Base=%s, N=%s)\n" % (Self.Type, Self.Base, Self.N)

class DISPLAY(Object) :
    def __init__(Self, Canvas, Base=[0,0]) :
        Object.__init__(Self, Canvas, Base)
        Self.Type = 'DISPLAY'
        Self.ID = "%s%05d" % (Self.Type, NextID())
        Self.Size = (80, 110)
        Self.Decode = {0: (1, 1, 1, 1, 1, 1, 0), 1: (0, 1 ,1, 0, 0, 0, 0), 2: (1, 1, 0, 1, 1, 0, 1), \
                       3: (1, 1, 1, 1, 0, 0, 1), 4: (0, 1, 1, 0, 0, 1, 1), 5: (1, 0, 1, 1, 0, 1, 1), \
                       6: (1, 0, 1, 1, 1, 1, 1), 7: (1, 1, 1, 0, 0, 0, 0), 8: (1, 1, 1, 1, 1, 1, 1), \
                       9: (1, 1, 1, 0, 0, 1, 1), 10: (1, 1, 1, 0, 1, 1, 1), 11: (0, 0, 1, 1, 1, 1, 1), \
                       12: (1, 0, 0, 1, 1, 1, 0), 13: (0, 1, 1, 1, 1, 0, 1), 14: (1, 0, 0, 1, 1, 1, 1), \
                       15: (1, 0, 0, 0, 1, 1, 1), -1: (0, 0, 0, 0, 0, 0, 0)}
        Self.Segments = ((30, 10, 60, 20), (60, 20, 70, 50), (60, 60, 70, 90), (30, 90, 60, 100), \
                         (20, 60, 30, 90), (20, 20, 30, 50), (30, 50, 60, 60))
        Self.Digit = -1
        Self.Terminals = (Terminal(Self, [0,20], [10,20]),  # 0: I0
                          Terminal(Self, [0,40], [10,40]),  # 1: I1
                          Terminal(Self, [0,60], [10,60]),  # 2: I2
                          Terminal(Self, [0,80], [10,80]))  # 3: I3

    def Simulate(Self) :
        """ This routine simulates a decoded seven segement display. If any of the
        inputs are not zero or one, the display is blank. Otherwise, a hexidecimal
        digit equal to the input value is displayed. """
        if Self.Terminals[0].Node.State > 1 or \
           Self.Terminals[1].Node.State > 1 or \
           Self.Terminals[2].Node.State > 1 or \
           Self.Terminals[3].Node.State > 1 :               # if I0-I3 = bad state
            Self.Canvas.master.Menu.Indicators["BadInput"].Update(True)
            Self.Digit = -1
        else :
            Self.Digit = Self.Terminals[0].Node.State + Self.Terminals[1].Node.State * 2 + \
                         Self.Terminals[2].Node.State * 4 + Self.Terminals[3].Node.State * 8
        return []

    def Draw_Digit(Self, Digit) :
        """ This routine draws a decoded digit based on a four bit binary number.
        using a seven segment display. _1 is resevred for a blank display. """
        X, Y = Self.Base
        X0 = X1 = Y0 = Y1 = 0
        SegStatus = Self.Decode[Digit]                      # get decoded segment values
        for Seg in xrange(7) :
            if SegStatus[Seg] == 1 :
                Color = "red"
            else :
                Color = "white"
            X0, Y0, X1, Y1 = Self.Segments[Seg]
            Self.Canvas.Create_Scaled_Rectangle(X+X0, Y+Y0, X+X1, Y+Y1, fill=Color, outline="gray", tag=Self.ID)

    def Draw(Self) :
        """ This routine draws a TOGGLE. """
        X, Y = Self.Base
        ID = Self.ID
        Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+80, Y+00, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+110, X+80, Y+110, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+10, Y+00, X+10, Y+110, tag=ID)
        Self.Canvas.Create_Scaled_Line(X+80, Y+00, X+80, Y+110, tag=ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+20, text='I0', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+40, text='I1', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+60, text='I2', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        Self.Canvas.Create_Scaled_Text(X+13, Y+80, text='I3', anchor='w', font=Self.Canvas.Font, tag=Self.ID)
        Self.Draw_Digit(Self.Digit)
        for Terminal in Self.Terminals :
            Terminal.Draw()

